/*----------------------------------------------------------------------
LFNDir.c
First Published	in PC Magazine,	US Edition, May	5, 1998
Written	by Rick	Knoblaugh
LFNDir performs	(most of) the equivalent functionality of the Windows 95
DIR command.
Bugs:
 - doesn't support FAT32 drives.
 - no caching scheme.
 - no function such as the Win 95 IFS manager UniToBCS
   service to convert to the base character set. To make it simple,
   it just pull out the first byte of each Unicode character.
Ver 1.0 -- Initial release
    1.1 fixes by Benny for TurboC
    1.1.1 minor changes by Benny for C->CPP
----------------------------------------------------------------------*/
#include <conio.h>
#include <string.h>
#include <dos.h>
#include <errno.h>
#include <stdio.h>
#include <io.h>
#include <ctype.h>
#include <stdlib.h>		// _psp
#include <bios.h>
#define	FAR
#ifdef __TURBOC__
#define	INT86X	int86x
#define	AX	ax
#define	AL	al
#define	AH	ah
#define	BX	bx
#define	CX	cx
#define	DX	dx
#define	DH	dh
#define	DL	dl
#define	CL	cl
#define	CH	ch
#define	DS	ds
#define	SI	si
#define	DI	di
#define	ES	es
#define	PSP	_psp
#endif
#include "lfnequ.h"
#include "lfnstruc.h"

void            display_entry();
// Program variables----------------------------------------------------
struct device_params drive_parms;       //(Benny)

struct disk_io  drive_io;
struct find_context last_find;	//keep track of where left off in searches

struct find_context hold_context;
struct media_info media_id;
struct rwblock  rwblk;
struct SREGS    segregs;
union REGS      inregs, outregs, regs;

unsigned char   dir_of_buf[MAX_PATH_LEN];
unsigned char   entry_format_buf[MAX_PATH_LEN + 80];
unsigned char   lfn_buf[MAX_PATH_LEN];
unsigned char   fat_sec_buf[512];
unsigned char   file_sys_buf[255];
unsigned char   hold_prev_buf[512];
unsigned char   hold_sec_buf[512];
unsigned char   prev_root_buf[512];
unsigned char   prev_sub_buf[512];
unsigned char   root_dir_buf[512];
unsigned char   sub_dir_buf[512];
unsigned char   work_dir_path[MAX_PATH_LEN];
unsigned char   search_attribute;
unsigned char  *dir_path_ptr;
unsigned char   last_error;
unsigned        bits_per_fat_entry;
unsigned        byte_offset;
unsigned        dir_drive;
unsigned        fat_sec;
unsigned        levels_in_path;
unsigned        number_of_rows;
unsigned        rem_bits;
unsigned        rem_byte_offset;
unsigned        row_count;
unsigned long   bit_offset;
unsigned long   hi_cluster;
unsigned long   lo_cluster;
unsigned long   start_root_dir;
unsigned long   start_of_data;
unsigned long   tot_bits;
unsigned long   tot_bytes_used;
unsigned long   tot_number_dirs;
unsigned long   tot_number_files;
const char      err_access_msg[] = "Error accessing drive\n";
const char      err_unknown_msg[] = "Unknown drive\n";
const char      err_notrdy_msg[] = "Drive is not ready\n";
const char      err_fat32_msg[] = "LFNDir doesn't support FAT32 drives\n";
const char      err_compres_msg[] = "LFNDir doesn't support compressed drives\n";
const char      wild_string[] = "*.*";
const char      dir_string[] = "   <DIR>      ";
const char      noname_string[] = "           ";
const char      has_no_string[] = "has no label";
const char      no_path_msg[] = "Path not found\n";
const char      no_file_msg[] = "File not found\n";
unsigned char   serial_present;

unsigned char   bare_flag = FALSE;
unsigned char   pause_flag = FALSE;
unsigned char   lower_flag = FALSE;
unsigned char   help_flag = FALSE;
unsigned char   attrib_flag = FALSE;
unsigned char   not_attrib_flag = FALSE;

unsigned char   attrib_tab[] =
{'D', SUBDIR_ATTRIB,
 'H', HIDDEN_ATTRIB,
 'S', SYSTEM_ATTRIB,
 'R', READONLY_ATTRIB,
 'A', ARCHIVE_ATTRIB,
 0, 0
};
struct switch_engine switch_table[] =
{
    {BARE_DISP_SWITCH, 0, 0, 0, FALSE, 0, &bare_flag, 0},
    {LOWER_SWITCH, 0, 0, 0, FALSE, 0, &lower_flag, 0},
    {PAUSE_SWITCH, 0, 0, 0, FALSE, 0, &pause_flag, 0},
    {HELP_SWITCH1, 0, 0, 0, FALSE, 0, &help_flag, 0},
    {HELP_SWITCH2, 0, 0, 0, FALSE, 0, &help_flag, 0},
    {ATTRIB_SWITCH, 5, ':', attrib_tab, TRUE, &not_attrib_flag, &attrib_flag, &search_attribute},
};
#define	num_switch_entries (sizeof(switch_table) / (sizeof(struct switch_engine)))
struct ext_fcb  fcb_data =
{0xff, "\0\0\0\0", 8, 0, "??????????", 0, 0, 0, 0, 0, 0, 0, 0};

// --------------------------------------------------------------------
//Note: These functions which access sectors for the root directory, sub-
//      directory, and FAT, could possibly be enhanced to use a
//      caching scheme.
unsigned        get_sec(unsigned char *sec_buf_ptr, unsigned long sec_num,
			unsigned scount, unsigned gsdrive)
{
    struct disk_io FAR *rw_ptr;
    rw_ptr = &drive_io;
    rw_ptr->dio_startsec = sec_num;
    rw_ptr->dio_numsecs = scount;	//number of sectors to read

    rw_ptr->dio_buf_ptr = sec_buf_ptr;

    inregs.h.AL = gsdrive - 1;	//0=a:, 1=b:, etc.

    inregs.x.CX = 0xffff;	//indicate to get info from struct

    segregs.DS = FP_SEG((void FAR *) rw_ptr);
    inregs.x.BX = FP_OFF((void FAR *) rw_ptr);
    INT86X(0x25, &inregs, &outregs, &segregs);
    last_error = outregs.h.AL;
    return (outregs.x.cflag);
}

unsigned        get_root_dir_sec(unsigned long root_sec)
{
    //store previous buffer (may need for retrieving LFNs)
    memmove(prev_root_buf, root_dir_buf, sizeof(prev_root_buf));
    return (get_sec(root_dir_buf, root_sec + start_root_dir, 1, dir_drive));
}

unsigned        get_sub_dir_sec(unsigned long dir_sec)
{
    //store previous buffer (may need for retrieving LFNs)
    memmove(prev_sub_buf, sub_dir_buf, sizeof(prev_sub_buf));
    return (get_sec(sub_dir_buf, dir_sec, 1, dir_drive));
}

unsigned        get_fat_sec(unsigned long fat_sec)
{
    return (get_sec(fat_sec_buf, fat_sec + 1, 1, dir_drive));
}


unsigned        see_if_star(unsigned char *work_ptr)
{
    while (*work_ptr != '\0' && *work_ptr != BACK_SLASH) {
	if (*work_ptr == '*')
	    return (TRUE);
	++work_ptr;
    }
    return (FALSE);
}

unsigned        see_if_wild(unsigned char *work_ptr)
{
    while (*work_ptr != '\0' && *work_ptr != BACK_SLASH) {
	if (*work_ptr == '?' || *work_ptr == '*')
	    return (TRUE);
	++work_ptr;
    }
    return (FALSE);
}


void            store_dir_of()
{
    unsigned        dir_of_len;
    //store lfn for "directory of" display
    //Go place the lfn (if it exists --otherwise place the formatted 8.3)
    //into the "directory of" hold buffer
    dir_of_buf[last_find.last_level_off++] = BACK_SLASH;
//    lfn_or_formatted83(&dir_of_buf[last_find.last_level_off]);
    dir_of_len = strlen(&dir_of_buf[last_find.last_level_off]);
    //note location for placing next portion
    last_find.last_level_off += dir_of_len;
    dir_of_buf[last_find.last_level_off] = 0;
}

/*-----------------------------------------------------------
| detect8_3 - called to	determine if a path element is	    |
|	      is 8.3 style or an LFN.			    |
|	Enter:						    |
|	     name_ptr =	ptr to path element to inspect	    |
|	 Exit:						    |
|	   Return TRUE if this is an 8.3 type file or	    |
|	   subdirectory	name.  Return FALSE if it's an LFN. |
------------------------------------------------------------*/
unsigned        detect8_3(unsigned char *name_ptr)
{
    unsigned        x;

    //LFNs can actually have periods anywhere in the filename --even
    //as the first character.
    if (*name_ptr == DOT)
	return (FALSE);
    for (x = 0; x < 8; x++, ++name_ptr) {
	//spaces only allowed in LFNs
	if (*name_ptr == SPACE)
	    return (FALSE);
	if (*name_ptr == '\0' || *name_ptr == BACK_SLASH ||
	    *name_ptr == '*')
	    return (TRUE);
	if (*name_ptr == DOT)
	    break;
    }
    if (x == 8) {
	if (*name_ptr != DOT &&
	    *name_ptr != '\0' && *name_ptr != BACK_SLASH)
	    return (FALSE);
	if (*name_ptr == '\0' || *name_ptr == BACK_SLASH)
	    return (TRUE);
    }
    ++name_ptr;			//get past the period

    for (x = 0; x < 4; x++, ++name_ptr) {
	//spaces only allowed in LFNs
	if (*name_ptr == SPACE)
	    return (FALSE);
	if (*name_ptr == '\0' || *name_ptr == BACK_SLASH ||
	    *name_ptr == '*')
	    return (TRUE);
	if (*name_ptr == DOT)
	    return (FALSE);
    }
    return (FALSE);
}

/*-----------------------------------------------------------
| retrieve_lfn - Found an alias.  Now attempt to retrieve   |
|		 the associated	LFN and	store it.  LFNs	    |
|		 must directly precede the alias (this	    |
|		 can be	in previous directory sector if	    |
|		 necessary).				    |
|	Enter:						    |
|	  sec_buf_ptr =	ptr to start of	directory sector    |
|	 find_con_ptr =	ptr to context of find operations   |
|			important info in structure:	    |
|			prev_sec_ptr (ptr to previous dir   |
|			sector contents)		    |
|			fc_entry_num (the last searched	    |
|			entry in the current dir sector)    |
|	lfn_store_ptr =	ptr to area where LFN should be	    |
|			placed.				    |
|	 Exit:						    |
|	   Store the LFN.  If not found, store 0	    |
------------------------------------------------------------*/
void            retrieve_lfn(unsigned char *sec_buf_ptr, struct find_context *find_con_ptr,
			     unsigned char *lfn_store_ptr)
{
    unsigned        x, z;
    struct lfn_dir_layout *lfn_dir_ptr;
    unsigned char   entry_start_search;
    unsigned char  *start_lfn_ptr;
    unsigned        safety_count = 0;

    start_lfn_ptr = lfn_store_ptr;
    *lfn_store_ptr = '\0';	//default to not found

    if (find_con_ptr->fc_entry_num > 0) {
	entry_start_search = find_con_ptr->fc_entry_num - 1;
	lfn_dir_ptr = (struct lfn_dir_layout *)
	    (sec_buf_ptr + (entry_start_search * LEN_DIR_ENTRY));
    } else {
	if (find_con_ptr->prev_sec_ptr == 0)
	    return;
	lfn_dir_ptr = (struct lfn_dir_layout *)
	    (find_con_ptr->prev_sec_ptr + ((ENTRIES_PER_SEC - 1) * LEN_DIR_ENTRY));
	entry_start_search = ENTRIES_PER_SEC - 1;
    }

    while (1) {
	//Note: LFNs are stored in Unicode. It would be nice to have
	//a function such as the Win 95 IFS manager UniToBCS
	//service to easily convert it to the base characer set. Since it's
	//not that simple in vanilla DOS, just pull out the first byte
	//of each Unicode character.

	for (x = 0; x <= entry_start_search; x++) {
	    //just keep copying until a null is reached (or until an
	    //entry that isn't an lfn is reached --if the lfn fits
	    //exactly in an entry, there's no null).
	    ++safety_count;
	    if (((lfn_dir_ptr->lfn_dir_atrrib & LFN_ATTRIB) != LFN_ATTRIB)
		|| safety_count > MAX_ENTRIES_FOR_LFN) {
		*lfn_store_ptr = '\0';	//terminate it

		return;
	    }
	    for (z = 0; z < sizeof(lfn_dir_ptr->lfn_name_area1); z++, z++) {
		*lfn_store_ptr++ = lfn_dir_ptr->lfn_name_area1[z];
		if (!lfn_dir_ptr->lfn_name_area1[z])
		    return;
	    }
	    for (z = 0; z < sizeof(lfn_dir_ptr->lfn_name_area2); z++, z++) {
		*lfn_store_ptr++ = lfn_dir_ptr->lfn_name_area2[z];
		if (!lfn_dir_ptr->lfn_name_area2[z])
		    return;
	    }
	    for (z = 0; z < sizeof(lfn_dir_ptr->lfn_name_area3); z++, z++) {
		*lfn_store_ptr++ = lfn_dir_ptr->lfn_name_area3[z];
		if (!lfn_dir_ptr->lfn_name_area3[z])
		    return;
	    }

	    --lfn_dir_ptr;	//move back in sector buffer to preceding entry


	}			//for x -- entries in the directory sector

	//if still haven't found all of the LFN, we need to need to
	//retrieve the rest from the previous directory sector (if it
	//exists)
	if (*find_con_ptr->prev_sec_ptr == '\0') {
	    *start_lfn_ptr = '\0';	//disregard any previous stuff

	    return;
	} else {
	    lfn_dir_ptr = (struct lfn_dir_layout *)
		(find_con_ptr->prev_sec_ptr + ((ENTRIES_PER_SEC - 1) * LEN_DIR_ENTRY));
	    entry_start_search = ENTRIES_PER_SEC - 1;
	}

    }				//while haven't retrieved all of LFN (or hit sanity checks --if
    //for some reason, things are corrupted)

}

/*-----------------------------------------------------------
| compare_lfn -	helper function	to find_match. Do all the   |
|		same stuff, except compare specifically	on  |
|		the LFN.				    |
|	Enter:						    |
|	  sec_buf_ptr =	ptr to start of	directory sector    |
|	 find_con_ptr =	ptr to context of find operations   |
|	lfn_store_ptr =	ptr to area where LFN has been	    |
|			placed.				    |
|	file_name_ptr =	ptr to element we're looking for.   |
|			This should be an LFN.		    |
|			have this attribute.		    |
|	   spec_flags =	caller can place values	in here	to  |
|			handle special situations.  These   |
|			take care of things that would	    |
|			really be a hassle otherwise.	    |
|			If specfied (non zero) values can   |
|			be:				    |
|			Value	 Meaning		    |
|			-----	 --------		    |
|	     NO_ATTRIB_IF_SUB	 Don't check attribute if   |
|				 item turns out	to be a	    |
|				 directory.		    |
|	   GEN_WILD_EXTENSION	 If there's no extension    |
|				 on a file being searched   |
|				 for, match it as if	    |
|				 there's a ".*" at end.     |
|	 Exit:						    |
|	   Return TRUE if match	found			    |
|	   Find	context	updated				    |
------------------------------------------------------------*/
unsigned        compare_lfn(unsigned char *sec_buf_ptr,
			    struct find_context *find_con_ptr,
		 unsigned char *file_name_ptr, unsigned char *lfn_store_ptr)
{
    unsigned        match_flag = TRUE;

    //no compare if LFN doesn't exist
    if (*lfn_store_ptr == 0)
	return (FALSE);

    for (;; ++file_name_ptr, ++lfn_store_ptr) {
	if ((*file_name_ptr == '\0' || *file_name_ptr == BACK_SLASH)
	    && *lfn_store_ptr == '\0')
	    break;
	//above, if hit the end of both, it was a match
	//if caller has specified in flags that we should treat
	//the file as if it had ".*", see if this will
	//enable it to match and tell caller this occurred.
	if (find_con_ptr->fc_spec_flags & GEN_WILD_EXTENSION) {
	    if (*file_name_ptr == '\0' && *lfn_store_ptr == DOT) {
		//treat as if filespec had ".*" at the end and
		//indicate to caller that this has been done
		find_con_ptr->fc_spec_flags |= FM_SET_WILD_EXTENSION;
		break;
	    }
	}
	//if have: 123456789.xxx.yyy and enter: 123456789.
	//don't want files that contain anything beyond the period
	if (*file_name_ptr == DOT && *(file_name_ptr + 1) == '\0')
	    if (*lfn_store_ptr == '\0')
		break;
	//if have: 123456789 and enter: 123456789.*
	//also want to match
	if (*file_name_ptr == DOT && *(file_name_ptr + 1) == '*')
	    if (*lfn_store_ptr == '\0')
		break;
	if (*file_name_ptr == '*')
	    break;
	if (*file_name_ptr == '?' && *lfn_store_ptr != '\0')
	    continue;
	//below, if hit the end of either, it's no match
	if ((*file_name_ptr == '\0' || *file_name_ptr == BACK_SLASH)
	    || *lfn_store_ptr == '\0') {
	    match_flag = FALSE;
	    break;
	}
	if ((toupper(*file_name_ptr)) != (toupper(*lfn_store_ptr))) {
	    match_flag = FALSE;	//no match

	    break;
	}
    }				//while checking lfn

    if (match_flag) {
	if ((find_con_ptr->dir_entry_ptr->file_attrib & SUBDIR_ATTRIB)
	    == SUBDIR_ATTRIB)
	    ++find_con_ptr->number_dirs;
	else {
	    ++find_con_ptr->number_files;
	    find_con_ptr->bytes_used += find_con_ptr->dir_entry_ptr->file_size;
	}
	//matched on LFN. We may or may not have made the
	//match happen by making the extension wild. LFNs
	//can contain so many period characters...just set
	//flag so that upper level will try additional finds
	//using the wild card extension feature.
	if (find_con_ptr->fc_spec_flags & GEN_WILD_EXTENSION)
	    find_con_ptr->fc_spec_flags |= FM_SET_WILD_EXTENSION;
    }				//if match

    sec_buf_ptr = sec_buf_ptr;  // (Benny)

    return (match_flag);
}

/*-----------------------------------------------------------
| find_match - Attempt to find matching	directory entry.    |
|	Enter:						    |
|	  sec_buf_ptr =	ptr to start of	directory sector    |
|	 find_con_ptr =	ptr to context of find operations   |
|	lfn_store_ptr =	ptr to area where LFN should be	    |
|			placed.				    |
|	file_name_ptr =	ptr to element we're looking for.   |
|			This may be a subdirectory or a	    |
|			file.  Drive letter and	or "\"	    |
|			have been removed from first part   |
|			of string.  If may contain wildcard |
|			chars (? or *).	 It may	be an LFN   |
|			or 8.3 string.	We search until	    |
|			hit terminating	"\" or 0.	    |
|	  find_attrib =	only match on dir entries that	    |
|			have this attribute.		    |
|	   spec_flags =	caller can place values	in here	to  |
|			handle special situations.  These   |
|			take care of things that would	    |
|			really be a hassle otherwise.	    |
|			If specfied (non zero) values can   |
|			be:				    |
|			Value	 Meaning		    |
|			-----	 --------		    |
|	     NO_ATTRIB_IF_SUB	 Don't check attribute if   |
|				 item turns out	to be a	    |
|				 directory.		    |
|	   GEN_WILD_EXTENSION	 If there's no extension    |
|				 on a file being searched   |
|				 for, match it as if	    |
|				 there's a ".*" at end.     |
|	 Exit:						    |
|	   Return TRUE if match	found and store	LFN	    |
|	   Find	context	updated				    |
------------------------------------------------------------*/
unsigned        find_match(unsigned char *sec_buf_ptr,
			   struct find_context *find_con_ptr,
		 unsigned char *lfn_store_ptr, unsigned char *file_name_ptr,
			   unsigned char find_attrib, unsigned spec_flags)
{
    unsigned        isit_8_3 = FALSE;
    unsigned        no_extension;
    unsigned        match_flag;
    unsigned        entry_to_start;
    struct dir_layout *dir_entry_ptr;
    unsigned        x, y;
    unsigned char  *start_file_name;
    unsigned char   work_attrib;
    start_file_name = file_name_ptr;
    isit_8_3 = detect8_3(file_name_ptr);
    entry_to_start = find_con_ptr->fc_entry_num;
    dir_entry_ptr = (struct dir_layout *) (sec_buf_ptr + (LEN_DIR_ENTRY * find_con_ptr->fc_entry_num));
    for (x = entry_to_start; x < ENTRIES_PER_SEC; match_flag = FALSE, x++, ++dir_entry_ptr) {
	find_con_ptr->fc_entry_num = x;
	//note the entry last searched
	find_con_ptr->dir_entry_ptr = dir_entry_ptr;	//and the ptr to it's data

	file_name_ptr = start_file_name;	//beginning of filename for searching
	//Locate only the alias first --skip LFN entries

	if ((dir_entry_ptr->file_attrib & LFN_ATTRIB) == LFN_ATTRIB)
	    continue;

	//not interested in volume labels
	if ((dir_entry_ptr->file_attrib & VOLUME_ATTRIB) == VOLUME_ATTRIB)
	    continue;

	//
	//The following handles special request processing by caller.  If
	//the special handling code of NO_ATTRIB_IF_SUB is specified, don't
	//apply attribute criterion to the item. Callers usually set this
	//prior to searching for the last path element.
	//If the item is a file, then the attribute criterion should
	//be applied. If it's a sub directory, caller will want to match on
	//this, then generate another search of "*.*"
	if (!((find_con_ptr->fc_spec_flags & NO_ATTRIB_IF_SUB) &&
	      (dir_entry_ptr->file_attrib & SUBDIR_ATTRIB))) {
	    if (((find_attrib == NORMAL_ATTRIB) && (dir_entry_ptr->file_attrib & 0x0f) > READONLY_ATTRIB))
		continue;

	    //Get only only entries that match desired attribute

	    //check for NOT first (it will be the NOT of attributes, so upper
	    //bits being on indicate '-' was specified)
	    if (find_attrib & 0x80) {
		work_attrib = ~find_attrib;
		if ((dir_entry_ptr->file_attrib & work_attrib) == work_attrib)
		    continue;
	    } else if ((dir_entry_ptr->file_attrib & find_attrib) != find_attrib)
		continue;
	}			//if it's not a subdirectory (when NO_ATTRIB_IF_SUB option being used)

	if (dir_entry_ptr->filename[0] == 0 ||
	    dir_entry_ptr->filename[0] == ERASE_FLAG)
	    continue;

	retrieve_lfn(sec_buf_ptr, find_con_ptr, lfn_store_ptr);
	match_flag = TRUE;	//default to having found a match

	no_extension = FALSE;

	//if item being searched for is an LFN, go use LFN compare logic
	if (!(isit_8_3)) {
	    if (compare_lfn(sec_buf_ptr, find_con_ptr,
			    start_file_name, lfn_store_ptr))
		return (TRUE);
	    else
		continue;
	}
	for (y = 0; y < 8; y++, ++file_name_ptr)	//main portion of file name
	 {
	    if (*file_name_ptr == '?')
		continue;	//if wild card char
	    //Note that normal DOS only really deals with wildcards for
	    //file names. For example:
	    //          dir subnam*       will find the subdir
	    //but, if you do this:  dir subnam*\*.*   --it won't find the
	    //subdirectory
	    //lfndir allows wildcards to be specified as part of subdirectories --
	    //in such cases, it will find the first match.

	    if (*file_name_ptr == '*') {
		++file_name_ptr;
		break;
	    }
	    if ((*file_name_ptr == BACK_SLASH || *file_name_ptr == '\0')
		&& dir_entry_ptr->filename[y] == SPACE)
		break;
	    //determine if hit '.' and main portion dir name at end
	    if (y > 0 && *file_name_ptr == DOT && dir_entry_ptr->filename[y] == SPACE)
		break;
	    if ((toupper(*file_name_ptr)) != (toupper(dir_entry_ptr->filename[y]))) {
		match_flag = FALSE;	//no match
		break;
	    }
	}	//for Y - each char in main portion of file name

	//if no match, check the 8.3 item being searched for to see if
	//wildcard was specifed. It's possible that even though it
	//looks like an 8.3, it will only match on the LFN. For example,
	//if the filename is: this123456789
	//the alias will be: this12~1
	//if we're looking for this123*
	//it looks like an 8.3, but will only match on the LFN
	if (match_flag == FALSE) {
	    if (see_if_star(start_file_name)) {
		if (compare_lfn(sec_buf_ptr, find_con_ptr,
				start_file_name, lfn_store_ptr))
		    return (TRUE);
		else
		    continue;
	    }
	    continue;		//go to next dir entry and try again

	}
	//if element ends with '\\', we're looking for a subdirectory
	if (*file_name_ptr == BACK_SLASH && (dir_entry_ptr->file_attrib & SUBDIR_ATTRIB)
	    != SUBDIR_ATTRIB)
	    continue;

	if (*file_name_ptr == BACK_SLASH || *file_name_ptr == '\0')
	    no_extension = TRUE;

	/*do some checking to handle the extension situation
	   LFNDIR *        Gives you everything (as does LFNDIR *.*  --
	   caller takes care of this one by setting GEN_WILD_EXTENSION )
	   LFNDIR *.       Gives you only files with no extensions
	   also        LFNDIR FOO.     Gives you only "FOO" as main with no extensions
	 */
	if (no_extension == FALSE) {
	    if (*file_name_ptr == DOT) {
		++file_name_ptr;	//advance to extension
		//if file spec ends with a dot, don't want to get
		//matching files that contain extensions

		if (*file_name_ptr == BACK_SLASH || *file_name_ptr == '\0')
		    find_con_ptr->fc_spec_flags &= ~GEN_WILD_EXTENSION;
	    }
	    if (*file_name_ptr == BACK_SLASH || *file_name_ptr == '\0')
		no_extension = TRUE;
	}
	for (y = 0; y < 3; y++)	//extension portion of name
	 {
	    //if there must be no extension to qualify as a match
	    if (no_extension) {
		if (dir_entry_ptr->fileext[y] != SPACE) {
		    if (find_con_ptr->fc_spec_flags & GEN_WILD_EXTENSION) {
			//treat as if filespec had ".*" at the end and
			//indicate to caller that this has been done
			find_con_ptr->fc_spec_flags |= FM_SET_WILD_EXTENSION;
			break;
		    }
		    match_flag = FALSE;		//no match

		    break;
		}
		continue;
	    }			//if there's no extension

	    if (*file_name_ptr == '*') {
		++file_name_ptr;
		break;
	    }
	    //if hit end, the remaining part of the directory
	    //entry must be blank.
	    if (*file_name_ptr == BACK_SLASH || *file_name_ptr == '\0') {
		if (dir_entry_ptr->fileext[y] == SPACE)
		    break;
		match_flag = FALSE;
		break;
	    }
	    //if not single char wild card --these will simply fall through
	    if (*file_name_ptr != '?') {
		if ((toupper(*file_name_ptr)) != (toupper(dir_entry_ptr->fileext[y]))) {
		    match_flag = FALSE;		//no match

		    break;
		}
	    }
	    ++file_name_ptr;
	}			// for y - each character in the extension portion of dir entry

	if (match_flag == FALSE)
	    continue;
	//if element ends with '\\', we're looking for a subdirectory
	if (*file_name_ptr == BACK_SLASH && (dir_entry_ptr->file_attrib & SUBDIR_ATTRIB)
	    != SUBDIR_ATTRIB)
	    continue;
	else
	    break;
    }				// for X - each directory entry in the sector being inspected

    if (match_flag) {
	//If option for specified to match on files regardless of
	//extension. Indicate that at least one match was found
	//even though there was no extension at all --this will
	//in turn cause upper level to make multiple calls to
	//look for further matches with varying extensions.
	if ((find_con_ptr->fc_spec_flags & GEN_WILD_EXTENSION) &&
	    no_extension)
	    find_con_ptr->fc_spec_flags |= FM_SET_WILD_EXTENSION;

    }
    //add to the total number of directories or files found (and number
    //of total bytes occupied by the files)
    if (match_flag) {
	if ((dir_entry_ptr->file_attrib & SUBDIR_ATTRIB)
	    == SUBDIR_ATTRIB)
	    ++find_con_ptr->number_dirs;
	else {
	    ++find_con_ptr->number_files;
	    find_con_ptr->bytes_used += dir_entry_ptr->file_size;
	}
    }
    spec_flags = spec_flags;    //(Benny)

    return (match_flag);
}

/*-----------------------------------------------------------
| find_in_root - search	root directory for a specified	    |
|		 entry.					    |
|	Enter:						    |
|   fc_ptr->fc_root_sec	= sector of root dir to	start	    |
|			  search.			    |
| fc_ptr->fc_root_entry	= entry	within current dir sector   |
|			  to begin this	search.		    |
|	       path_ptr	= ptr to item being searched for.   |
|	       lfnb_ptr	= ptr to buffer	in which to place   |
|			  long filename	counterpart to	    |
|			  this entry.			    |
|	    file_attrib	= attribute to also be matched.	    |
|	 Exit:						    |
|	       return TRUE if found			    |
|	       fc_error	in find_context	struct holds any    |
|	       error code from the disk	I/O		    |
------------------------------------------------------------*/
unsigned        find_in_root(struct find_context *fc_ptr,
			     unsigned char *lfnb_ptr,
			 unsigned char *path_ptr, unsigned char file_attrib)
{
    unsigned        x;
    for (x = fc_ptr->fc_root_sec; x < (drive_parms.dp_bpb_de / ENTRIES_PER_SEC); x++) {
	if (fc_ptr->fc_root_entry == 0)
	    //if starting at first entry in sector
	{
	    //if not at the first sector, store previous
	    if (x != 0)
		fc_ptr->prev_sec_ptr = prev_root_buf;
	    //go get the next root dir sector --return if       error
	    if (get_root_dir_sec(x))
		return (FALSE);
	}
	//Tell find_match which dir entry at which to start search.
	//(Separate hold values are maintained for root and for
	//currently being searched entry.)
	fc_ptr->fc_entry_num = fc_ptr->fc_root_entry;
	fc_ptr->fc_root_sec = x;
	if (find_match(root_dir_buf, fc_ptr, lfnb_ptr,
		       path_ptr, file_attrib, fc_ptr->fc_spec_flags)) {
	    //note where found
	    fc_ptr->fc_root_entry = fc_ptr->fc_entry_num + 1;
	    if (fc_ptr->fc_root_entry == ENTRIES_PER_SEC) {
		fc_ptr->fc_root_entry = 0;
		++fc_ptr->fc_root_sec;
	    }
	    return (TRUE);
	}
	fc_ptr->fc_root_entry = 0;
	//can't go further if disk inaccessible
	if (fc_ptr->fc_error)
	    break;
    }				//for x - all root dir sectors

    return (FALSE);
}

/*-----------------------------------------------------------
| get_clusters - Given a cluster number, go to the FAT and  |
|		 find the next cluster in the chain.	    |
|	Enter:						    |
| bits_per_fat_entry = number of bits per entry	(either	    |
|		       12 or 16	--FAT 32 not supported)	    |
|	 Exit:						    |
|	       return next cluster (or zero if at end)	    |
------------------------------------------------------------*/
unsigned        get_clusters(unsigned long cluster)
{
    tot_bits = (unsigned long) (cluster * (unsigned long) bits_per_fat_entry);
    fat_sec = (tot_bits / BITS_PER_SEC);
    rem_bits = (tot_bits % BITS_PER_SEC);
    bit_offset = rem_bits;
    //return 0 if I/O error
    if (get_fat_sec(fat_sec))
	return (0);
    byte_offset = bit_offset / 8;
    rem_byte_offset = bit_offset % 8;
    lo_cluster = (unsigned long) ((unsigned char) fat_sec_buf[byte_offset]);
    if (rem_byte_offset)
	lo_cluster = lo_cluster >> 4;
    ++byte_offset;
    if ((BITS_PER_SEC - tot_bits) < bits_per_fat_entry) {
	++fat_sec;
	if (get_fat_sec(fat_sec))
	    return (0);
	byte_offset = 0;
    }
    hi_cluster = (unsigned long) ((unsigned char) fat_sec_buf[byte_offset]);

    if (rem_byte_offset == 0) {
	if (bits_per_fat_entry == 12) {
	    hi_cluster = hi_cluster & 0x0f;
	    hi_cluster = hi_cluster << 8;
	} else
	    hi_cluster = hi_cluster << 8;
    } else {
	if (bits_per_fat_entry == 12)
	    hi_cluster = hi_cluster << 4;
    }
    lo_cluster = lo_cluster | hi_cluster;

    //if at end of chain, return 0
    if ((bits_per_fat_entry == 12 && lo_cluster == 0x0fff) ||
	(bits_per_fat_entry == 16 && lo_cluster == 0x0ffff))
	return (0);
    return ((unsigned) lo_cluster);
}


/*-----------------------------------------------------------
| find_in_sub -	search subdirectory for	a specified entry.  |
|	Enter:						    |
|	 fc_ptr->fc_cluster = cluster to start search.	    |
| fc_ptr->fc_cluster_offset = sector offset within cluster  |
|			      to begin this search.	    |
|	     fc_ptr->fc_sec = sector to	start search (note  |
|			      if entry offset == 0 and	    |
|			      cluster offset ==	0, we	    |
|			      are just starting	search of   |
|			      this cluster, so this sector  |
|			      value is calculated).	    |
|      fc_ptr->fc_entry_num = directory	entry within sector |
|			      to start search.		    |
|	   (other parms	same as	root dir search)	    |
|	 Exit:						    |
|	       return TRUE if found			    |
|	       fc_error	in find_context	struct holds any    |
|	       error code from the disk	I/O		    |
------------------------------------------------------------*/
unsigned        find_in_sub(struct find_context *fc_ptr,
			    unsigned char *lfnb_ptr,
			 unsigned char *path_ptr, unsigned char file_attrib)
{
    unsigned        clust_off, match_flag;

    while (TRUE) {
	//First, check to see if last search reached end of cluster. If it
	//did, go read FAT and get next entry.
	if (fc_ptr->fc_cluster_offset == drive_parms.dp_bpb_au) {
	    //get next in chain, return 0 if end of chain
	    if ((fc_ptr->fc_cluster = get_clusters(fc_ptr->fc_cluster)) == 0)
		break;
	    fc_ptr->fc_cluster_offset = 0;
	    fc_ptr->fc_entry_num = 0;
	}
	for (clust_off = fc_ptr->fc_cluster_offset;
	     clust_off < drive_parms.dp_bpb_au; ++clust_off) {
	    fc_ptr->fc_cluster_offset = clust_off;
	    if (clust_off == 0 && fc_ptr->fc_entry_num == 0)
		fc_ptr->fc_sec = (((fc_ptr->fc_cluster - 2) * drive_parms.dp_bpb_au)
				  + start_of_data) - 1;
	    if (!(fc_ptr->fc_entry_num))	//if no offset into a sector
	     {
		if (clust_off != 0)
		    fc_ptr->prev_sec_ptr = prev_sub_buf;
		//go get the next dir sector --return if error
		if (get_sub_dir_sec(fc_ptr->fc_sec))
		    return (FALSE);
	    }
	    match_flag = find_match(sub_dir_buf, fc_ptr, lfnb_ptr,
			      path_ptr, file_attrib, fc_ptr->fc_spec_flags);
	    ++fc_ptr->fc_entry_num;
	    if (fc_ptr->fc_entry_num == ENTRIES_PER_SEC) {
		fc_ptr->fc_entry_num = 0;
		++fc_ptr->fc_sec;
		fc_ptr->fc_cluster_offset = clust_off + 1;
	    }
	    if (match_flag)
		return (TRUE);
	    //can't go further if disk inaccessible
	    if (fc_ptr->fc_error)
		return (FALSE);
	}			//for clust_off --all sectors within current cluster

    }				//while --stay until end of cluster chain reached

    return (FALSE);
}

/*-----------------------------------------------------------
| detect_subdir	- search root or other directory to	    |
|		  determine whether the	specfied entry	    |
|		  is a directory. If found, contexts are    |
|		  updated, as callers will want	to pick	    |
|		  up from the found directory. If a	    |
|		  directory isn't found, contexts are       |
|		  preserved.				    |
|	Enter:						    |
|   Parmameters	are the	same as	the find_in_root type	    |
|   functions, with one	additional parameter:		    |
|	 dir_for_search	= 0 indicates root, 1 indicates	    |
|			  current directory.		    |
|	 Exit:						    |
|	       return TRUE if found			    |
------------------------------------------------------------*/
unsigned        detect_subdir(struct find_context *fc_ptr,
			      unsigned char *lfnb_ptr,
			 unsigned char *path_ptr, unsigned char file_attrib,
			      unsigned dir_for_search)
{
    unsigned        match_flag;
    //save contents of previous sector buffer (this is used for retrieving
    //LFNs
    memmove(hold_prev_buf, fc_ptr->prev_sec_ptr, sizeof(hold_prev_buf));
    //save entire context structure. If subdirectory isn't found, we'll
    //restore everything so that other searches can start from this
    //location
    memmove(&hold_context, fc_ptr, sizeof(struct find_context));
    fc_ptr->fc_spec_flags = NO_ATTRIB_IF_SUB;
    if (dir_for_search == 0) {
	memmove(hold_sec_buf, root_dir_buf, sizeof(hold_sec_buf));
	if (!(match_flag = find_in_root(fc_ptr, lfn_buf, path_ptr, file_attrib)))
	    memmove(root_dir_buf, hold_sec_buf, sizeof(root_dir_buf));

    } else {
	memmove(hold_sec_buf, sub_dir_buf, sizeof(sub_dir_buf));
	if (!(match_flag = find_in_sub(fc_ptr, lfn_buf, path_ptr, file_attrib)))
	    memmove(sub_dir_buf, hold_sec_buf, sizeof(sub_dir_buf));
    }
    if (match_flag) {
	if ((fc_ptr->dir_entry_ptr->file_attrib & SUBDIR_ATTRIB)
	    == SUBDIR_ATTRIB)
	    return (TRUE);
    }
    memmove(fc_ptr, &hold_context, sizeof(struct find_context));
    memmove(fc_ptr->prev_sec_ptr, hold_prev_buf, sizeof(hold_prev_buf));
    lfnb_ptr = lfnb_ptr;        // (Benny)

    return (FALSE);
}

/*-----------------------------------------------------------
| get_volume_info - Use	Get Device Parameters IOCTL to	    |
|		    retrieve characteristics of	drive.	    |
|		    Also, given	the information, calculate  |
|		    and	store away the sector numbers where |
|		    the	root directory and data	areas begin.|
|	Enter:						    |
|	     drive_no =	number of drive	to check	    |
|	 Exit:						    |
|      start_root_dir =	first sector for root directory	    |
|	start_of_data =	first sector of	data area	    |
|	   Return TRUE if error				    |
|	   If successful, drive_parms structure	is filled   |
|	   in with the drive information.		    |
------------------------------------------------------------*/
//struct  device_params;
unsigned        get_volume_info(unsigned drive_no)
{
#define	device_parms device_params
    struct device_params FAR *dev_ptr;  // (Benny) wooohaaa!

    dev_ptr = (struct device_parms FAR *) &drive_parms;

    drive_parms.dp_spec_func = 1;	//retrieve for current drive

    inregs.x.AX = 0x440d;
    inregs.x.BX = drive_no;
    inregs.h.CH = MAJ_CODE;	//must be 08

    inregs.h.CL = GET_DEVICE_PARMS;	//get device parms function

    segregs.DS = FP_SEG((void FAR *) dev_ptr);
    inregs.x.DX = FP_OFF((void FAR *) dev_ptr);
    INT86X(0x21, &inregs, &outregs, &segregs);

    start_root_dir = (drive_parms.dp_bpb_fs * drive_parms.dp_bpb_nf) + 1;
    start_of_data = start_root_dir + (drive_parms.dp_bpb_de / ENTRIES_PER_SEC) + 1;
    return (outregs.x.cflag);
}

unsigned        determine_compressed(unsigned drive_no)
{
    char            volume_string[] = "A:\\";
    void FAR       *pass_ptr;
    volume_string[0] = (char) (drive_no + 'A' - 1);
    pass_ptr = (void FAR *) volume_string;
    inregs.x.AX = GET_VOLUME_INFO;
    inregs.x.CX = sizeof(file_sys_buf);		//max length of buffer

    segregs.DS = FP_SEG(pass_ptr);
    inregs.x.DX = FP_OFF(pass_ptr);
    pass_ptr = (void FAR *) file_sys_buf;
    segregs.ES = FP_SEG((void FAR *) pass_ptr);
    inregs.x.DI = FP_OFF((void FAR *) pass_ptr);
    INT86X(0x21, &inregs, &outregs, &segregs);

    //This call may not be supported (could actually be running on
    //pure DOS)
    if (outregs.x.cflag)
	return (FALSE);
    //BX=0x8000 if volume is compressed
    return (outregs.x.BX & 0x8000);

}

/*-----------------------------------------------------------
| determine_fat_size - Find file system	type (12 or 16 bit).|
|		       Use Get Media ID	function to get	    |
|		       this information. This also retrieves|
|		       the volume label	and serial number.  |
|	Enter:						    |
|    drive_no =	drive number (a=1, b=2,	etc.)		    |
|	 Exit:						    |
|    Number of bits (12	or 16) returned.  Zero returned	    |
|    if	it's not a valid file system type.                  |
|    Serial number and volume label stored in media_id	    |
|    structure.						    |
|    serial_present = TRUE if call processed ok				      |
------------------------------------------------------------*/
unsigned        determine_fat_size(unsigned drive_no)
{
    struct media_info FAR *media_ptr;
    struct ext_fcb FAR *fcb_ptr;
    struct psp FAR *psp_ptr;
    unsigned long   num_sectors;
    unsigned        fat_size;
    unsigned        x;

    serial_present = TRUE;	//default to serial number found

    media_ptr = (struct media_info FAR *) &media_id;
    fcb_ptr = (struct ext_fcb FAR *) &fcb_data;
    inregs.x.AX = 0x440d;
    inregs.x.BX = drive_no;
    inregs.h.CH = MAJ_CODE;	//must be 08

    inregs.h.CL = GET_MEDIA_ID;
    segregs.DS = FP_SEG((void FAR *) media_ptr);
    inregs.x.DX = FP_OFF((void FAR *) media_ptr);
    INT86X(0x21, &inregs, &outregs, &segregs);

    //This call shouldn't fail. However, on older floppies where there is
    //not volume label, serial number, or indication of the file system,
    //it will fail. Thus, attempt alternate method of retrieving
    //the volume label and determine and return the FAT size.
    if (outregs.x.cflag) {
	serial_present = FALSE;	//no serial number present

	fcb_data.ext_filename[10] = '?';
	fcb_data.ext_driveid = drive_no;
	inregs.h.AH = FIND_FIRST_FILE;
	segregs.DS = FP_SEG((void FAR *) fcb_ptr);
	inregs.x.DX = FP_OFF((void FAR *) fcb_ptr);
	INT86X(0x21, &inregs, &outregs, &segregs);
	if (outregs.x.cflag)
	    return (0);
        {                       //(Benny)

	    unsigned        stg = PSP;
	    FP_SEG(psp_ptr) = stg;
	}
	FP_OFF(psp_ptr) = 0;

	fcb_ptr = (struct ext_fcb FAR *) &psp_ptr->psp_defdta;

	for (x = 0; x < sizeof(fcb_ptr->ext_filename); x++)
	    media_id.vol_label[x] = fcb_ptr->ext_filename[x];

	num_sectors = (unsigned long) drive_parms.dp_bpb_ts;
	if (drive_parms.dp_bpb_ts == 0)
	    num_sectors = drive_parms.dp_bpb_huge_sec;

	fat_size = 12;
	if ((num_sectors / drive_parms.dp_bpb_au) > 4094)
	    fat_size = 16;

	return (fat_size);

    }
    //looking for "FAT12" or "FAT16"
    if (media_id.file_sys_type[0] != 'F' || media_id.file_sys_type[3] !=
	'1')
	return (0);
    if (media_id.file_sys_type[4] != '2' && media_id.file_sys_type[4] !=
	'6')
	return (0);
    return ((media_id.file_sys_type[4] - '0') + 10);
}

/*-----------------------------------------------------------
| format_long -	want to	format an unsigned long	so that	    |
|		there are commas:  xx,xxx,xxx,xxx	    |
|	Enter:						    |
| form_buf_ptr = ptr the buffer	to hold	the formatted	    |
|		 number.				    |
|     form_num = number	to format			    |
|	 Exit:						    |
|	       buffer is filled				    |
------------------------------------------------------------*/
void            format_long(unsigned char *form_buf_ptr,
			    unsigned long form_num)
{
    unsigned char   work_form_buf[20];
    int             x;
    unsigned        digit_count;
    unsigned        len_needed;
    unsigned        len_buf;

    sprintf(work_form_buf, "%lu\0", form_num);
    len_buf = strlen(work_form_buf);
    len_needed = len_buf;
    len_needed += ((len_needed - 1) / 3);	//space needed for commas

    form_buf_ptr += len_needed;
    //zero terminator at end
    *form_buf_ptr-- = 0;
    digit_count = 1;
    //start building at end and work back
    for (x = len_buf - 1; x >= 0; x--, ++digit_count) {
	*form_buf_ptr-- = work_form_buf[x];
	if (x - 1 >= 0) {
	    if (digit_count == ((digit_count / 3) * 3))
		*form_buf_ptr-- = ',';	//add a comma

	}
    }
}

/*-----------------------------------------------------------
| levels_in_path - determine how many levels to	traverse.   |
|	Enter:						    |
|	 dir_path_ptr =	ptr to path string		    |
|	 Exit:						    |
|	       return the number			    |
------------------------------------------------------------*/
unsigned        find_levels(char *dir_path_ptr)
{
    unsigned        level_count = 0;

    do {
	if (*dir_path_ptr == BACK_SLASH)
	    ++level_count;
    } while (*dir_path_ptr++);
    return (level_count);
}

/*-----------------------------------------------------------
| do_dir - Main	processing for carrying	out DIR.	    |
|	Enter:						    |
|   dir_path_ptr = specifier on	which to perform the DIR.   |
|		   Note:  this has already been	"massaged"; |
|		   drive letter	has been removed;  any	    |
|		   periods indicating use current dir or    |
|		   parent have been replaced with	    |
|		   appropriate directory.  In these cases   |
|		   when	it could be determined that we're   |
|		   dealing with	a directory, "*.*" has	    |
|		   been	placed at the end of the string.    |
| levels_in_path = number of levels to traverse.	    |
|	 Exit:						    |
|	       return TRUE if error accessing disk	    |
------------------------------------------------------------*/
unsigned        do_dir()
{
    unsigned        list_all = FALSE;
    unsigned char  *work_ptr;
    unsigned        x;
    unsigned        match_flag;
    unsigned        match_subdir;
    unsigned char   attrib_to_use;
    ++dir_path_ptr;		//get past "\" (something is guaranteed to be there)


/*
#define NORMAL_ATTRIB 0
#define READONLY_ATTRIB 0x01
#define HIDDEN_ATTRIB 0x02
#define SYSTEM_ATTRIB 0x04
#define VOLUME_ATTRIB 0x08*/

    work_ptr = dir_path_ptr;

    //determine if there are wildcards
    list_all = see_if_wild(dir_path_ptr);
//    list_all = TRUE; // (Benny) added

    match_subdir = match_flag = FALSE;

    //Use normal attributes until we're at the last path element. Otherwise,
    //if user has specified an attribute like "hidden" like this:
    //lfndir \foo1\foo2\*.* /AH
    //We don't want to apply the hidden attribute criterion until the first
    //path elements have been matched.
    attrib_to_use = NORMAL_ATTRIB;
    if (levels_in_path == 1) {

#define MY_ATTRIB	HIDDEN_ATTRIB
//(READONLY_ATTRIB + HIDDEN_ATTRIB + SYSTEM_ATTRIB)

	attrib_to_use = search_attribute;
//      attrib_to_use = MY_ATTRIB; //(Benny)
	//last one, apply user attribute conditions
	if (!(list_all))	//no wildcards specified
	 {
	    //At the last element of the path, and there were no wild
	    //cards specified. Besides looking for what could be just
	    //a plain ol' file, there are also two other scenarios to
	    //be handled:  (1) the item could be file with no extension (this
	    //can also be an LFN with no extension --e.g.,
	    //foo123456789 in  which case you would want to find it
	    //as well as: foo1223456789.xxx.morestuff),
	    //(2) the item could be a subdirectory in which case you wouldn't
	    //want to try to find any other files regardless of a common
	    //main portion with various other extensions.
	    //There's no other way around it, before we attempt to look for
	    //files that may match up (and have other extensions), it must
	    //first be determined if this is an actual directory. The call
	    //below checks for this.
	    match_subdir = detect_subdir(&last_find, lfn_buf,
					 dir_path_ptr, attrib_to_use, 0);

	    //For the case where no extension has been supplied, tell low-level
	    //if it can't find a match, to generate a ".*" to the filespec. This
	    //saves us a ton of work to let that level deal with it. It in
	    //turn will indicate that it's done this by setting the
	    //FM_SET_WILD_EXTENSION bit in fc_spec_flags.

	}			//if no wildcards
	//This not only takes care of getting foo.exe when user
	//enters: lfndir foo    but it also takes care of giving
	//all files when the user enters: lfndir *

	last_find.fc_spec_flags = GEN_WILD_EXTENSION;
    }				//if only in the root (levels_in_path == 1)

    else
	last_find.fc_spec_flags |= NO_ATTRIB_IF_SUB;

    if (!match_subdir)
	match_flag = find_in_root(&last_find, lfn_buf, dir_path_ptr, attrib_to_use);
    if (match_flag || match_subdir) {
	//can't go further if disk inaccessible
	if (last_find.fc_error)
	    return (TRUE);

	last_find.fc_path_found = TRUE;		//so far, path is found
	last_find.fc_file_found = TRUE;		//so far, files found
	//if only 1 level in path and no wildcards, we're done (if it's
	//not a subdirectory --in which case, we need to go get all the
	//files in there)

	if (levels_in_path == 1 && !(list_all)) {
	    if ((last_find.dir_entry_ptr->file_attrib & SUBDIR_ATTRIB)
		== SUBDIR_ATTRIB) {
		//basically turn "\subdir" into "\subdir\*.*"
		++levels_in_path;
		work_ptr = dir_path_ptr + strlen(dir_path_ptr);
		*work_ptr++ = BACK_SLASH;
		memmove(work_ptr, wild_string, sizeof(wild_string) + 1);
	    } else		//not a subdir, we've found the only file
	     {
		//If low level match engine (in response to our setting
		//GEN_WILD_EXTENSION) detected that the last item searched for
		//had no extension, and was able to find a match by tacking on a ".*"
		//We'll want to fall through and loop through all directory entries
		//just as if user had dynamically specified:   foo.*
		if (last_find.fc_spec_flags & FM_SET_WILD_EXTENSION)
		    list_all = TRUE;
		else {
		    display_entry();
		    return (FALSE);
		}
	    }			//else --it's not a subdirectory
	}			//if one level and no wildcards

	//Dealing with more than 1 level (e.g., lfndir \foo1\foo2)  --must
	//do a sanity check to ensure that the entry matched on is
	//actually a subdirectory.
	if (levels_in_path > 1) {
	    if ((last_find.dir_entry_ptr->file_attrib & SUBDIR_ATTRIB)
		!= SUBDIR_ATTRIB) {
		last_find.fc_path_found = FALSE;	//path not found
		return (FALSE);
	    }
	}
    }				//if entry was found

    else {
	if (levels_in_path == 1)
	    last_find.fc_file_found = FALSE;
	else
	    last_find.fc_path_found = FALSE;
	return (FALSE);		//entry not found
    }

    //if listing all files that appear only in the root
    // (dealing with wildcards)
    if (levels_in_path == 1) {
	do {
	    if (last_find.fc_error)
		return (TRUE);
	    display_entry();
	}
	while (find_in_root(&last_find, lfn_buf, dir_path_ptr, attrib_to_use));

	return (FALSE);		//done

    }
    //if get here, we've found the first subdirectory in the root
    //now, using the cluster specified in directory entry to locate
    //the start of the subdirectory, we'll go search it
    list_all = FALSE;
    last_find.number_dirs = 0;	//only count dirs in final dir

    attrib_to_use = NORMAL_ATTRIB;
    last_find.fc_spec_flags = 0;

    for (x = 1; x < levels_in_path; x++) {
	while (*dir_path_ptr++ != BACK_SLASH);	//advance to next element
	match_subdir = match_flag = FALSE;
	last_find.fc_cluster = last_find.dir_entry_ptr->starting_cluster;
	last_find.fc_cluster_offset = 0;
	last_find.fc_entry_num = 0;
	last_find.prev_sec_ptr = 0;
	//store the directory for "directory of" display
	store_dir_of();

	if (x == levels_in_path - 1) {
	    list_all = see_if_wild(dir_path_ptr);
//          list_all = TRUE; // (Benny) added
	    last_find.number_dirs = 0;	//only count dirs in final dir

	    //last one, apply user attribute conditions
            attrib_to_use = search_attribute; //(Benny)
//          attrib_to_use =     MY_ATTRIB; //(Benny)


	    if (!list_all) {
		match_subdir = detect_subdir(&last_find, lfn_buf,
					     dir_path_ptr, attrib_to_use, 1);
	    }
	    last_find.fc_spec_flags = GEN_WILD_EXTENSION;
	} else
	    last_find.fc_spec_flags |= NO_ATTRIB_IF_SUB;

	if (!match_subdir)
	    match_flag = find_in_sub(&last_find, lfn_buf, dir_path_ptr,
				     attrib_to_use);

	match_flag |= match_subdir;

	//can't go further if disk inaccessible
	if (last_find.fc_error)
	    return (TRUE);

	//if not a match and not at end of specified, path wasn't found
	//if at end and not found, it's file not found.
	if (!(match_flag)) {
	    if (x != levels_in_path - 1)
		last_find.fc_path_found = FALSE;	//path not found

	    else
		last_find.fc_file_found = FALSE;	//file not found

	    return (FALSE);	//not found

	}
	if (x != levels_in_path - 1)	//if not at last element of path
	 {
	    if ((last_find.dir_entry_ptr->file_attrib & SUBDIR_ATTRIB)
		!= SUBDIR_ATTRIB) {
		last_find.fc_path_found = FALSE;	//path not found

		return (FALSE);
	    }
	} else {
	    //Reached the last one.  Check to see if it's a directory (also
	    //check to see if this is only 1 level beyond root --if so,
	    //we've already done this)
	    if (((last_find.dir_entry_ptr->file_attrib & SUBDIR_ATTRIB)
		 == SUBDIR_ATTRIB) && !(list_all)) {
		++levels_in_path;
		work_ptr = dir_path_ptr + strlen(dir_path_ptr);
		*work_ptr++ = BACK_SLASH;
		memmove(work_ptr, wild_string, sizeof(wild_string) + 1);
	    }
	    //If low level match engine (in response to our setting
	    //GEN_WILD_EXTENSION) detected that the last item searched for
	    //had no extension, and was able to find a match by tacking on
	    // a ".*"
	    //We'll want to fall through and loop through all directory
	    // entries just as if user had dynamically specified foo.*
	    if (last_find.fc_spec_flags & FM_SET_WILD_EXTENSION)
		list_all = TRUE;
	}
    }				//for x - each element of the path

    if (!list_all) {
	display_entry();
	return (FALSE);
    }
    do {
	if (last_find.fc_error)
	    return (TRUE);
	display_entry();
    } while (find_in_sub(&last_find, lfn_buf, dir_path_ptr, attrib_to_use));
    return (FALSE);

}

void            init_context()
{
    last_find.prev_sec_ptr = 0;
    last_find.fc_root_sec = 0;
    last_find.fc_root_entry = 0;
    last_find.fc_sec = 0;
    last_find.fc_entry_num = 0;
    last_find.fc_cluster = 0;
    last_find.fc_cluster_offset = 0;
    last_find.number_dirs = 0;
    last_find.number_files = 0;
    last_find.bytes_used = 0;
    last_find.fc_error = FALSE;
    last_find.fc_path_found = TRUE;
    last_find.fc_file_found = TRUE;
    last_find.last_level_off = 0;
    last_find.fc_spec_flags = 0;
    //init "directory of" buffer to root
    dir_of_buf[0] = BACK_SLASH;
}

#define OK 	1
#define FAILED  0


short           StoreDirLine()
{

#define vfat  			last_find.dir_entry_ptr
#define l_short_filename 	filename
#define l_ext			file_ext
#define l_attrib 		file_attrib
#define l_size 			file_size
#define l_year			fd_yr
#define l_month			fd_mo
#define	l_day			fd_day
#define	l_hour			ft_hrs
#define l_min			ft_mins
#define vfat_long_filename	lfn_buf

   // is it ".\0"?
   // current dir
   if (vfat->l_short_filename[0] == '.'
	&& vfat->l_short_filename[1] != '.')
	{ // do nothing
   // label?
   } else if (vfat->l_attrib & FA_LABEL) {
    // is it not ".\0"?
   } else {
	ctl++;

#ifdef OVERFLOWCHECK
	if (ctl >= FILES_MAX) {
	    ctl--;
	    fileoverflow = '+';
	    return 1;
	}
#endif
	if (ctl > loggedmax) {
	    logged[ctl] = new FILEBLOCK;
	    if (!logged[ctl]) {
		ctl--;
		fileoverflow = '+';
		return 1;
	    }
	    loggedmax++;
	}
	logged[ctl]->dir = 0x20;
	logged[ctl]->ext[0] = logged[ctl]->ext[1] =
	    logged[ctl]->ext[2] = 0;
#define UP_MARK          UPMARK
#define DOWN_MARK	 DOWNMARK

	// is it a dir UP?
	if (strncmp(vfat->l_short_filename, "..", 2) == SAME) {
	    if (vfat->l_short_filename[8] != ' ')
		sprintf(logged[ctl]->name,
			"%c%.8s.%.3s", UP_MARK, vfat->l_short_filename,
			vfat->l_short_filename + 8);
	    else
		sprintf(logged[ctl]->name,
			"%c%.8s    ", UP_MARK, vfat->l_short_filename,
			vfat->l_short_filename + 8);
	    FlagTopDir = YES;
	    logged[ctl]->dir = UP_MARK;
	    logged[ctl]->ext[0] = logged[ctl]->ext[1] =
		logged[ctl]->ext[2] = UP_MARK;
	    // is it a dir DOWN?
	} else if (vfat->l_attrib & FA_DIREC) {
	    if (vfat->l_short_filename[8] != ' ')
		sprintf(logged[ctl]->name,
			"%c%.8s.%.3s", DOWN_MARK, vfat->l_short_filename,
			vfat->l_short_filename + 8);
	    else
		sprintf(logged[ctl]->name,
			"%c%.8s    ", DOWN_MARK, vfat->l_short_filename,
			vfat->l_short_filename + 8);
	    logged[ctl]->dir = DOWN_MARK;
	    logged[ctl]->ext[0] = logged[ctl]->ext[1] =
		logged[ctl]->ext[2] = DOWN_MARK;
	    // is it a file?
	} else {
	    // NAMESIZE = 14 ABCDEFGH.EXT
	    // 012345678901
	    if (vfat->l_short_filename[8] != ' ') {
		char *ptr;
		ptr = strchr(vfat->l_short_filename, ' ');
		if (ptr) ptr[0] = 0;
		sprintf(logged[ctl]->name, "%.8s.%.3s",
			vfat->l_short_filename, vfat->l_short_filename + 8);
		}
	    else
		sprintf(logged[ctl]->name, "%.8s    ",
			vfat->l_short_filename, vfat->l_short_filename + 8);
	    memcpy(logged[ctl]->ext, vfat->l_short_filename + 8, 3);
	}
	totalsize += vfat->l_size;
	logged[ctl]->size = vfat->l_size;
	logged[ctl]->attrib = vfat->l_attrib;
	logged[ctl]->day = vfat->l_day;
	logged[ctl]->month = vfat->l_month;
	logged[ctl]->year = vfat->l_year;
	if (logged[ctl]->year >= 20)
	    logged[ctl]->year -= 20;
	else
	    logged[ctl]->year += 80;
	logged[ctl]->hour = vfat->l_hour;
	logged[ctl]->min = vfat->l_min;

	// does this helps? (sort bug)
	logged[ctl]->date =
	31*((vfat->l_year*12) + vfat->l_month) + vfat->l_day;
	logged[ctl]->time = vfat->l_hour*60 + vfat->l_min;

	if (lfn_buf[0] == '\0')	//if no lfn
	    logged[ctl]->LFN[0] = 0;
	else {
	    memcpy(logged[ctl]->LFN, vfat_long_filename, LASTLINE);
	    logged[ctl]->LFN[LASTLINE-1] = 0;
	    }

	logged[ctl]->key = 0;
//	memcpy(&logged[ctl]->key, &logged[ctl]->name, 8);
	if (SortType == BYEXT)
	    memcpy(&logged[ctl]->key, &logged[ctl]->ext, 3);
	else if (SortType == BYDATE) {
	    ulong           temp, tempkey;
	    tempkey =
		(ulong)
		logged[ctl]->year << 22
		+ logged[ctl]->month << 18
		+ logged[ctl]->day << 12
		+ logged[ctl]->hour << 6
		+ logged[ctl]->min;
	    logged[ctl]->key = tempkey;
	} else if (SortType == BYSIZE)
	    logged[ctl]->key = logged[ctl]->size;
#ifdef USEBIT
	logged[ctl]->key &= BITMASK;	// erase top bit

	if (logged[ctl]->dir == UP_MARK) {
	    logged[ctl]->key |= UPBIT;
	    logged[ctl]->key &= UPBITMASK;	// erase 2nd bit
	} else if (logged[ctl]->dir == DOWN_MARK)
	    logged[ctl]->key |= DOWNBIT;
	else
	    logged[ctl]->key |= TOPBIT;
#endif
    }
#ifdef DEBUG0
    printf("'%.11s' size=%i attrib=%i %i/%i/%i %i:%i\n",
	   vfat->l_short_filename,
	   vfat->l_size,
	   vfat->l_attrib,
	   vfat->l_day,
	   vfat->l_month,
	   vfat->l_year,
	   vfat->l_hour,
	   vfat->l_min
	);
    if (lfn_buf[0] != '\0')	//if lfn
	printf("LFN='%40s'\n", vfat_long_filename);
#endif

    return 0;
}

void            display_entry()
{
    StoreDirLine();
}


#ifdef DEBUG
int             ctt = 0;
#define DEBUGCTT	printf("ctt = %i, ",ctt++);
#else
#define DEBUGCTT
#endif

lfndir(char *file_spec)
{
    unsigned char  *user_path_ptr = file_spec;
    unsigned char   cur_dir_path[MAX_PATH_LEN];

    DEBUGCTT
    dir_drive = toupper(file_spec[0]) - 'A' + 1;

    // get filespec
    user_path_ptr += 2;		//get past drive letter and ':'

    strcpy(cur_dir_path, file_spec + 2);
#ifdef DEBUG0
    printf("#%i cur_dir_path= '%s'\n", dir_drive, cur_dir_path);                //(Benny)
#endif
    DEBUGCTT

    //Go load work_dir_path with specifier.  May simply copy what
    //user entered, or generate one from current path (e.g., if
    //"LFNDIR" (with no parms), generate "\curpath\*.*".
//    put_together_spec(user_path_ptr, cur_dir_path, work_dir_path);
#ifdef DEBUG0
    printf("?work_dir_path= '%s'\n", work_dir_path);
#endif
    strcpy(work_dir_path, cur_dir_path);
    if (strlen(work_dir_path) == 1)
	strcat(work_dir_path, "*.*\0");
//    else strcat(work_dir_path, "\*.*\0");
    dir_path_ptr = work_dir_path;

    // the only 3 printf errmsg
    //attempt to use int 25h (this will fail on FAT32 volumes)
    if (get_sec(fat_sec_buf, 1, 1, dir_drive)) {
	puts("fail 1: FAT32 found");
	printf((char *) &err_fat32_msg);
	return FAILED;
    }
    if (determine_compressed(dir_drive)) {
	puts("fail 2: compressed drive found");
	printf((char *) &err_compres_msg);
	return FAILED;
    }
    if (get_volume_info(dir_drive)) {
	puts("fail 3: volume info error");
	printf((char *) &err_access_msg);
	return FAILED;
    }
    DEBUGCTT
    //go determine FAT size
    //if I/O error, or unable to determine FAT size, can't continue
	if (!(bits_per_fat_entry = determine_fat_size(dir_drive))) {
	puts("fail 4: fat size calculation error");
	return FAILED;
    }
    //go determine how many levels deep we'll need to look
    levels_in_path = find_levels(work_dir_path);
    init_context();		//zero out history of sectors searched

    //go perform DIR processing and exit if error is returned
    if (do_dir()) {
	puts("fail 5: dir failure");
	return FAILED;
    }
    if (last_find.fc_path_found == FALSE) {
	puts("fail 6: path not found");
	return FAILED;
    }
    DEBUGCTT
	return OK;
}
